# IFDB Tag Manipulator
# by FLACRabbit
# version 0.2

fh = None

try:
	fh = open("config.txt")
except IOError:
	print("File config.txt not found, please create it and enter user details. Refer to README.txt for more information")
	quit()

content = fh.read()
fh.close()

lines = [x for x in content.split('\n') if x != '']

config = {}

for l in lines:
	sp = l.split(':')
	name = sp[0]
	value = sp[1]
	while value[0] == ' ':
		value = value[1:]
	if value != '':
		config[name] = value

USERNAME = ""
PASSWORD = ""

try:
	USERNAME = config['username']
except KeyError:
	print("Setting 'username' not found in config.txt. Please create it and enter your email address.")

try:
	PASSWORD = config['password']
except KeyError:
	print("Setting 'password' not found in config.txt. Please create it and enter your email address.")

from luxml import *
import urllib.request

SEARCH_BASE = 'https://ifdb.org/search?xml'

def char_unsafe(char):
	ascii = ord(char)
	if ascii >= ord('A') and ascii <= ord('z'):
		return False
	if ascii >= ord('0') and ascii <= ord('9'):
		return False
	return True

def to_hex(char):
	i = ord(char)
	first = 0
	while i > 16:
		i -= 16
		first += 1
	if first > 15:
		first = 15
	o = ''
	if first < 10:
		o += str(first)
	else:
		o += chr(first-10+ord('A'))
	if i < 10:
		o += str(i)
	else:
		o += chr(i-10+ord('A'))
	return o

def percentify(char):
	return f'%{to_hex(char)}'

def urlargument_process(value):
	if type(value) == int:
		return str(value)
	elif type(value) == str:
		s = ''
		for char in value:
			if char == ' ':
				s += '+'
			elif char_unsafe(char):
				s += percentify(char)
			else:
				s += char
		return s
	elif value:
		return urlargument_process(str(value))
	return value


class URLArgument:
	def __init__(self, name, value=None, process=True):
		self.name = name
		if process:
			self.value = urlargument_process(value)
		else:
			self.value = value
	def encoding(self):
		if self.value == None:
			return f'&{self.name}'
		else:
			return f'&{self.name}={self.value}'

def generate_url(base, arguments):
	url = base
	for a in arguments:
		url += a.encoding()
	return url

class IFDBGame:
	def __init__(self, name, tuid):
		self.name = name
		self.tuid = tuid
	def __eq__(self, other):
		# used for filtering
		try:
			return self.name == other.name and self.tuid == other.tuid
		except Exception:
			return False
	def __repr__(self):
		return f'IFDBGame({self.name}={self.tuid})'

def ifdb_search(s):
	u = generate_url(SEARCH_BASE, [URLArgument('game'), URLArgument('searchfor', s), URLArgument("pg", "all"), URLArgument("showFlagged", 1)])
	handle = urllib.request.urlopen(u)
	c = handle.read()
	handle.close()

	tree = luxml_parse(c.decode('utf-8'))[1]
	gamesn = luxml_node_recursive_subnodes_matching(tree, 'games')[0]
	games = []
	for game in luxml_node_subnodes(gamesn):
		tuid = luxml_node_recursive_subnodes_matching(game, 'tuid')[0].contents[0]
		name = luxml_node_all_text(luxml_node_recursive_subnodes_matching(game, 'title')[0])
		games.append(IFDBGame(name, tuid))
	return games

def games_with_tag(tag_name):
	return ifdb_search(f'tag:"{tag_name}"')

def games_with_genre(genre_name):
	return ifdb_search(f'genre:"{genre_name}"')

def games_with_system(system_name):
	return ifdb_search(f'system:"{system_name}"')

def games_with_language(language_name):
	return ifdb_search(f'tag:"{language_name}"')

class IFDBGameTag:
	def __init__(self, game_tuid, name, assignment_count):
		self.game_tuid = game_tuid
		self.name = name
		self.assignment_count = assignment_count
	def __repr__(self):
		return f'IFDBGameTag({self.assignment_count}x {self.name} on {self.game_tuid})'

import datetime

def get_game_tuid_tags(tuid, mine_only=False):
	args = [URLArgument('id', tuid)]
	if mine_only:
		args += [URLArgument('mine_only', 'true')]
		args += [URLArgument('username', USERNAME, False)]
		args += [URLArgument('password', PASSWORD, False)]
	data = generate_url('', args)

	handle = urllib.request.urlopen('https://ifdb.org/gametags', data=data.encode('utf-8'))
	c = handle.read()
	handle.close()

	tree = luxml_parse(c.decode('utf-8'))[1]
	ts = datetime.datetime.now().strftime("%m%d%Y_%H%M%S")
	luxml_write_file([luxml_generate_prolog(), tree], f'tag_backup_{str(tuid)}_{ts}.xml')

	tags = []
	for tag_node in luxml_node_recursive_subnodes_matching(tree, 'tag'):
		tag_name = luxml_node_all_text(luxml_node_recursive_subnodes_matching(tag_node, 'name')[0])
		tag_name = tag_name.lower()
		tag_count = int(luxml_node_all_text(luxml_node_recursive_subnodes_matching(tag_node, 'tagcnt')[0]))
		tags.append(IFDBGameTag(tuid, tag_name, tag_count))

	return tags

def overwrite_game_tuid_tags(tuid, tags):
	taggame_args = [
		URLArgument('id', tuid),
		URLArgument('username', USERNAME, False),
		URLArgument('password', PASSWORD, False),
	]
	for i, t in enumerate(tags):
		taggame_args += [URLArgument(f't{i}', t.name)]

	data = generate_url('', taggame_args)

	handle = urllib.request.urlopen('https://ifdb.org/taggame?xml', data=data.encode('utf-8'))
	c = handle.read()
	handle.close()

	tree = luxml_parse(c.decode('utf-8'))[1]
	label = luxml_node_recursive_subnodes_matching(tree, 'label')[0]
	lt = luxml_node_all_text(label)
	if lt != 'Saved':
		raise Exception('Tag overwrite failed.')

def add_tag(tuid, tag_name):
	my_tags = get_game_tuid_tags(tuid, True)
	all_tags = get_game_tuid_tags(tuid, False)
	if any([x.name.lower() == tag_name for x in all_tags]):
		return False
	overwrite_game_tuid_tags(tuid, my_tags+[IFDBGameTag(tuid, tag_name, 1)])
	return True

def game_from_name(game_name):
	g = ifdb_search(game_name)[0]
	if g.name == game_name:
		return g
	else:
		return None

import sys

if '--help' in sys.argv or len(sys.argv) == 1:
	print("tm.py - the IFDB Tag Manipulator\n\t-sg [game title]: select game by title\n\t-st [tag name]: select all games with tag\n\t-sge [genre name]: select all games with genre\n\t-sl [language]: select all games with language\n\t-ss [system]: select all games with implementation system\n\t-at [tag name]: add tag to all selected games\n\t-rt [tag name]: remove tag from all selected games\n\t-u: disable confirmation prompt\n\nUse with no actions given to display a table of selected games.\nReplace the initial s of a selection argument with a f to make it into a filter argument.")
	quit()

arguments = sys.argv[1:]

abbreviations = {
	'-sg':'--select-game',
	'-st':'--select-tag',
	'-sge':'--select-genre',
	'-ss':'--select-system',
	'-sl':'--select-language',

	'-ft':'--filter-tag',
	'-fge':'--filter-genre',
	'-fs':'--filter-system',
	'-fl':'--filter-language',

	'-at':'--add-tag',
	'-rt':'--remove-tag',
	'-u':'--unsafe'
}

confirm = True

games = []
filter_list = None
actions = []
i = 0
while i < len(arguments):
	arg = arguments[i]
	if arg in abbreviations:
		arg = abbreviations[arg]
	if arg[:2] == '--':
		# TODO: refactor and fix code duplication
		if arg == '--unsafe':
			confirm = False
		elif arg == '--select-game':
			i += 1
			if i >= len(arguments):
				print('Error: missing game title after --select-game')
				quit()
			g = game_from_name(arguments[i])
			if g == None:
				print(f'Warning: no game found with title "{arguments[i]}"')
			else:
				games += [g]
		elif arg == '--select-tag':
			i += 1
			if i >= len(arguments):
				print('Error: missing tag name after --select-tag')
				quit()
			games += games_with_tag(arguments[i])
		elif arg == '--select-genre':
			i += 1
			if i >= len(arguments):
				print('Error: missing genre name after --select-genre')
				quit()
			games += games_with_genre(arguments[i])
		elif arg == '--select-system':
			i += 1
			if i >= len(arguments):
				print('Error: missing system name after --select-system')
				quit()
			games += games_with_system(arguments[i])
		elif arg == '--select-language':
			i += 1
			if i >= len(arguments):
				print('Error: missing language name after --select-language')
				quit()
			games += games_with_language(arguments[i])
		elif arg == '--filter-tag':
			i += 1
			if i >= len(arguments):
				print('Error: missing tag name after --filter-tag')
				quit()
			if filter_list == None:
				filter_list = games_with_tag(arguments[i])
			else:
				gwt = games_with_tag(arguments[i])
				filter_list = [x for x in filter_list if x in gwt]
		elif arg == '--filter-genre':
			i += 1
			if i >= len(arguments):
				print('Error: missing genre name after --filter-genre')
				quit()
			if filter_list == None:
				filter_list = games_with_genre(arguments[i])
			else:
				gwg = games_with_genre(arguments[i])
				filter_list = [x for x in filter_list if x in gwg]
		elif arg == '--filter-system':
			i += 1
			if i >= len(arguments):
				print('Error: missing system name after --filter-system')
				quit()
			if filter_list == None:
				filter_list = games_with_system(arguments[i])
			else:
				gws = games_with_system(arguments[i])
				filter_list = [x for x in filter_list if x in gws]
		elif arg == '--filter-language':
			i += 1
			if i >= len(arguments):
				print('Error: missing language name after --filter-system')
				quit()
			if filter_list == None:
				filter_list = games_with_language(arguments[i])
			else:
				gws = games_with_language(arguments[i])
				filter_list = [x for x in filter_list if x in gws]
		elif arg == '--add-tag':
			i += 1
			if i >= len(arguments):
				print('Error: missing tag name after --add-tag')
				quit()
			actions += [('tag', arguments[i])]
		elif arg == '--remove-tag':
			i += 1
			if i >= len(arguments):
				print('Error: missing tag name after --remove-tag')
				quit()
			actions += [('rtag', arguments[i])]
		else:
			print(f'Error: unrecognized long-form argument {arg}')
			quit()
	elif arg[0] == '-':
		print(f'Error: unrecognized short-form argument {arg}')
		quit()
	else:
		print(f'Error: unrecognized or incorrectly placed argument {arg}')
	i += 1

stripped_games = []

for g in games:
	if g not in stripped_games:
		stripped_games.append(g)

games = stripped_games

if filter_list != None:
	games = [x for x in games if x in filter_list]

if len(games) == 0:
	print('Error: no games selected')
	quit()

for a in actions:
	command = a[0]
	if command == 'tag':
		new_tag = a[1]
		if new_tag:
			for g in games:
				go = False
				if confirm == False:
					go = True
				else:
					go = input(f'Confirm tagging {g.name} with {new_tag}? ').lower() in ['y', 'yes', 'true', 't']
				if go:
					try:
						print(f'Tagging {g.name} with {new_tag}...', end='')
						result = add_tag(g.tuid, new_tag)
						if result == False:
							print(' Already assigned.')
						else:
							print(' Done.')
					except Exception:
						print(f' Error occured.')
	if command == 'rtag':
		new_tag = a[1]
		if new_tag:
			for g in games:
				tags = get_game_tuid_tags(g.tuid, True)
				if new_tag.lower() not in [x.name for x in tags]:
					continue
				go = False
				if confirm == False:
					go = True
				else:
					go = input(f'Confirm removing {new_tag} from {g.name}? ').lower() in ['y', 'yes', 'true', 't']
				if go:
					try:
						print(f'Removing {new_tag} from {g.name}...', end='')
						tags = [x for x in tags if x.name != new_tag]
						overwrite_game_tuid_tags(g.tuid, tags)
						print('Done.')
					except Exception:
						print(' Error occured.')

if len(actions) == 0:
	spacing = 0
	for game in games:
		if len(game.name) > spacing:
			spacing = len(game.name)
	spacing += 4
	s = f'Game'
	s += ' '*(spacing-4)
	s += f'TUID'
	print(s)
	print('-'*(spacing+16))
	for game in sorted(games, key=lambda x: x.name):
		s = f'{game.name}'
		s += ' '*(spacing-len(game.name))
		s += f'{game.tuid}'
		print(s)
